/* mpir helper routines
*/

#include "mpi.h"
#include "sbcnst.h"

MPIR_dump_rhandle(handle)
MPIR_RHANDLE handle;
{
    printf("  handle type = %d\n", handle.handle_type);
    printf("  source      = %d\n", handle.source);
    printf("  tag         = %d\n", handle.tag);
    printf("  persistence = %d\n", handle.persistence);
    printf("  completed   = %d\n", handle.completed);
    printf("  datatype    = "); MPIR_dump_datatype(handle.datatype);
}

MPIR_dump_shandle(handle)
MPIR_SHANDLE handle;
{
    printf("  handle type = %d\n", handle.handle_type);
    printf("  dest =        %d\n", handle.dest);
    printf("  tag =         %d\n", handle.tag);
    printf("  mode =        %d\n", handle.mode);
    printf("  persistence = %d\n", handle.persistence);
    printf("  completed   = %d\n", handle.completed);
    printf("  datatype    = "); MPIR_dump_datatype(handle.datatype);
}

MPIR_dump_queue(header)
MPIR_QHDR *header;
{
    MPIR_QEL *p;

    if (!header) return;
    printf("first = 0x%x, last = 0x%x, maxlen = %d, currlen = %d\n",
	   header->first, header->last, header->maxlen, header->currlen );
    p = header->first;
    while ( p )
    {
	switch ( p->qel_type )
	{
	  case MPIR_QSHANDLE:
	    printf("queued send handle:\n");
	    MPIR_dump_shandle( *((MPIR_SHANDLE *) p->ptr) );
	    break;
	  case MPIR_QRHANDLE:
	    printf("queued recv handle:\n");
	    MPIR_dump_rhandle( *((MPIR_RHANDLE *) p->ptr) );
	    break;
	}
	p = p->next;
    }
}

MPIR_enqueue(header, object, object_type)
MPIR_QHDR     *header;
void          *object;
MPIR_QEL_TYPE object_type;
{
    MPIR_QEL *p;

    header->currlen++;
    if (header->currlen > header->maxlen)
	header->maxlen++;

    p           = (MPIR_QEL *) SBalloc( MPIR_qels );
    p->ptr      = object;
    p->qel_type = object_type;

    if (header->first == NULL)
    {
	header->first = header->last = p;
	p->prev       = p->next      = NULL;
    }
    else
    {
	p->prev            = header->last;
	p->next            = NULL;
	header->last->next = p;
	header->last       = p;
    }
}

MPIR_dequeue(header, object)
MPIR_QHDR *header;
void      *object;
{
    MPIR_QEL *p;

    p = header->first;
    while ( p != NULL )
    {
	if ( p->ptr == object )
	    break;
    }
    if ( p == NULL )
	MPIR_error("MPIR_dequeue: not found");
    else
    {
	if ( p->next != NULL )
	    p->next->prev = p->prev;
	if ( p->prev != NULL )
	    p->prev->next = p->next;
	header->currlen--;
	printf("fix up header prts!\n");
    }	    
}


MPIR_dump_dte( dte, indent )
MPIR_DATATYPE *dte;
int           indent;
{
    int i;

    switch (dte->dte_type)
    {
      case MPIR_INT:
	MPIR_Tab( indent );
	printf( "int\n" );
	break;
      case MPIR_FLOAT:
	MPIR_Tab( indent );
	printf( "float\n" );
	break;
      case MPIR_DOUBLE:
	MPIR_Tab( indent );
	printf( "double\n" );
	break;
      case MPIR_BYTE:
	MPIR_Tab( indent );
	printf( "byte\n" );
	break;
      case MPIR_CHAR:
	MPIR_Tab( indent );
	printf( "char\n" );
	break;
      case MPIR_LONG:
	MPIR_Tab( indent );
	printf( "long\n" );
	break;
      case MPIR_SHORT:
	MPIR_Tab( indent );
	printf( "short\n" );
	break;
      case MPIR_CONTIG:
	MPIR_Tab( indent );
	printf( "contig, count = %d\n", dte->count );
	MPIR_dump_dte( dte->old_type, indent + 2 );
	break;
      case MPIR_VECTOR:
	MPIR_Tab( indent );
	printf( "vector, count = %d, stride = %d, blocklen = %d\n",
	       dte->count, dte->stride, dte->blocklen );
	MPIR_dump_dte( dte->old_type, indent + 2 );
	break;
      case MPIR_HVECTOR:
	MPIR_Tab( indent );
	printf( "hvector, count = %d, stride = %d, blocklen = %d\n",
	       dte->count, dte->stride, dte->blocklen );
	MPIR_dump_dte( dte->old_type, indent + 2 );
	break;
      case MPIR_INDEXED:
	MPIR_Tab( indent );
	printf( "indexed, count = %d\n", dte->count );
	MPIR_dump_dte( dte->old_type, indent + 2 );
	for ( i = 0; i < dte->count; i++)
	{
	    MPIR_Tab( indent + 4 );
	    printf("index = %d, blocklen = %d\n",
		   dte->indices[i], dte->blocklens[i] );
	}
	break;
      case MPIR_HINDEXED:
	MPIR_Tab( indent );
	printf( "hindexed, count = %d\n", dte->count );
	MPIR_dump_dte( dte->old_type, indent + 2 );
	for ( i = 0; i < dte->count; i++)
	{
	    MPIR_Tab( indent + 4 );
	    printf("index = %d, blocklen = %d\n",
		   dte->indices[i], dte->blocklens[i] );
	}
	break;
      case MPIR_STRUCT:
	MPIR_Tab( indent );
	printf( "struct, count = %d\n", dte->count );
	for ( i = 0; i < dte->count; i++)
	{
	    MPIR_Tab( indent + 2 );
	    printf("index = %d, blocklen = %d\n",
		   dte->indices[i], dte->blocklens[i] );
	    MPIR_dump_dte( dte->old_types[i], indent + 2 );
	}
	break;
    }
}

/* adds to singly-linked list of flat datatype elements, returns pointer to
   head and to "next" pointer in last element, for appending.  Updates 
   current displacement.
 */
MPIR_flatten_dte( dte, fdte, tailptr, disp )
MPIR_DATATYPE *dte;
MPIR_FDTEL    **fdte, ***tailptr;
int           *disp;
{
    int i;
    MPIR_FDTEL *p, **q, **r;

/*
    printf("entering flatten dte, dte type = %d, count = %d\n",
	   dte->dte_type, dte->count);
*/
    switch (dte->dte_type)
    {
      case MPIR_INT:
	p         = (MPIR_FDTEL *) SBalloc( MPIR_fdtels );
	p->disp   = *disp;
	p->type   =  MPIR_INT;
	*disp     += sizeof( MPIR_INT );
	*tailptr  = &(p->next);
	*fdte     = p;
	break;
      case MPIR_FLOAT:
	p         = (MPIR_FDTEL *) SBalloc( MPIR_fdtels );
	p->disp   = *disp;
	p->type   =  MPIR_FLOAT;
	*disp     += sizeof( MPIR_FLOAT );
	*tailptr  = &(p->next);
	*fdte     = p;
	break;
      case MPIR_DOUBLE:
	p         = (MPIR_FDTEL *) SBalloc( MPIR_fdtels );
	p->disp   = *disp;
	p->type   = MPIR_DOUBLE;
	*disp    += sizeof( MPIR_DOUBLE );
	*tailptr  = &(p->next);
	*fdte     = p;
	break;
      case MPIR_CONTIG:
	r = &p;
	for (i = 0; i < dte->count; i++)
	{
	    MPIR_flatten_dte(dte->old_type, r, &q, disp );
	    if ( i == 0 )
		*fdte = p;	/* remember the first one */
	    r = q;
	}
	*tailptr = r;
	break;
      default:
	printf("mpir_flatten not implemented yet for type %d\n",dte->dte_type);
    }
    return( MPI_OK );
}

MPIR_dump_flat_dte( fdte )
MPIR_FDTEL *fdte;
{
    MPIR_FDTEL *p;

    p = fdte;
    while ( p )
    {
	printf("(%d,%d),", p->type, p->disp );
	p = p->next;
    }
    printf("\n");
}

MPIR_Tab( n  )
int n;
{
    int i;

    for (i = 0; i < n; i++)
	putchar(' ');
}


MPIR_error( message )
char *message;
{
    printf("%s\n", message);
    exit(-1);
}
